"""
Wedata FeatureStoreClient Python实现
"""

from __future__ import annotations
from typing import Union, List, Dict, Optional, Any
from pyspark.sql import DataFrame, SparkSession
from pyspark.sql.streaming import StreamingQuery
from pyspark.sql.types import StructType

from feature_store.constants.constants import APPEND, DEFAULT_WRITE_STREAM_TRIGGER
from feature_store.entities.feature_function import FeatureFunction
from feature_store.entities.feature_lookup import FeatureLookup
from feature_store.entities.training_set import TrainingSet
from feature_store.feature_table_client.feature_table_client import FeatureTableClient
from feature_store.spark_client.spark_client import SparkClient
from feature_store.training_set_client.training_set_client import TrainingSetClient
from feature_store.utils.feature_utils import format_feature_lookups_and_functions


class FeatureStoreClient:
    """特征存储统一客户端，提供特征全生命周期管理能力"""

    def __init__(self, spark: SparkSession):
        """
        :param spark: 已初始化的SparkSession对象
        """
        self._spark = spark
        self._spark_client = SparkClient(spark)
        self._feature_table_client = FeatureTableClient(spark)

    def create_table(
            self,
            name: str,
            primary_keys: Union[str, List[str]],
            df: Optional[DataFrame] = None,
            *,
            timestamp_keys: Union[str, List[str], None] = None,
            partition_columns: Union[str, List[str], None] = None,
            schema: Optional[StructType] = None,
            description: Optional[str] = None,
            tags: Optional[Dict[str, str]] = None
    ):
        """
        创建特征表（支持批流数据写入）

        Args:
            name: 特征表全称（格式：<table>）
            primary_keys: 主键列名（支持复合主键）
            df: 初始数据（可选，用于推断schema）
            timestamp_keys: 时间戳键（用于时态特征）
            partition_columns: 分区列（优化存储查询）
            description: 业务描述
            tags: 业务标签

        Returns:
            FeatureTable实例

        Raises:
            ValueError: 当schema与数据不匹配时
        """

        return self._feature_table_client.create_table(
            name=name,
            primary_keys=primary_keys,
            df=df,
            timestamp_keys=timestamp_keys,
            partition_columns=partition_columns,
            schema=schema,
            description=description,
            tags=tags
        )


    def read_table(self, name: str) -> DataFrame:
        """
        读取特征表数据

        Args:
            name: 特征表名称

        Returns:
            DataFrame: 包含特征表数据的DataFrame对象
        """
        return self._feature_table_client.read_table(name)


    def drop_table(self, name: str) -> None:
        """
        删除特征表

        Args:
            name: 要删除的特征表名称

        Returns:
            None
        """
        return self._feature_table_client.drop_table(name)


    def create_training_set(
            self,
            df: DataFrame,
            feature_lookups: List[Union[FeatureLookup, FeatureFunction]],
            label: Union[str, List[str], None],
            exclude_columns: Optional[List[str]] = None,
            **kwargs,
    ) -> TrainingSet:

        """
        创建训练集

        Args:
            df: 基础数据
            feature_lookups: 特征查询列表
            label: 标签列名
            exclude_columns: 排除列名

        Returns:
            TrainingSet实例
        """

        if exclude_columns is None:
            exclude_columns = []

        features = feature_lookups
        del feature_lookups

        features = format_feature_lookups_and_functions(self._spark_client, features)
        # 创建TrainingSetClient实例
        training_set_client = TrainingSetClient(self._spark_client)
        return training_set_client.create_training_set_from_feature_lookups(
            df=df,
            feature_lookups=features,
            label=label,
            exclude_columns=exclude_columns,
            **kwargs
        )

    def write_table(
            self,
            name: str,
            df: DataFrame,
            mode: str = APPEND,
            checkpoint_location: Optional[str] = None,
            trigger: Dict[str, Any] = DEFAULT_WRITE_STREAM_TRIGGER,
    ) -> Optional[StreamingQuery]:

        """
        写入数据到特征表（支持批处理和流式处理）

        Args:
            name: 特征表名称
            df: 要写入的数据DataFrame
            mode: 写入模式（默认追加）
            checkpoint_location: 流式处理的检查点位置（可选）
            trigger: 流式处理触发器配置（默认使用系统预设）

        Returns:
            如果是流式写入返回StreamingQuery对象，否则返回None
        """

        return self._feature_table_client.write_table(
            name=name,
            df=df,
            mode=mode,
            checkpoint_location=checkpoint_location,
            trigger=trigger,
        )


from typing import Dict



class FeatureTable:
    """
    .. note::

       Aliases:`!databricks.feature_engineering.entities.feature_table.FeatureTable`, `!databricks.feature_store.entities.feature_table.FeatureTable`

    Value class describing one feature table.

    This will typically not be instantiated directly, instead the
    :meth:`create_table() <databricks.feature_engineering.client.FeatureEngineeringClient.create_table>`
    will create :class:`.FeatureTable` objects.
    """

    def __init__(
        self,
        name,
        table_id,
        description,
        primary_keys,
        partition_columns,
        features,
        creation_timestamp=None,
        online_stores=None,
        notebook_producers=None,
        job_producers=None,
        table_data_sources=None,
        path_data_sources=None,
        custom_data_sources=None,
        timestamp_keys=None,
        tags=None,
    ):
        """Initialize a FeatureTable object."""
        self.name = name
        self.table_id = table_id
        self.description = description
        self.primary_keys = primary_keys
        self.partition_columns = partition_columns
        self.features = features
        self.creation_timestamp = creation_timestamp
        self.online_stores = online_stores if online_stores is not None else []
        self.notebook_producers = (
            notebook_producers if notebook_producers is not None else []
        )
        self.job_producers = job_producers if job_producers is not None else []
        self.table_data_sources = (
            table_data_sources if table_data_sources is not None else []
        )
        self.path_data_sources = (
            path_data_sources if path_data_sources is not None else []
        )
        self.custom_data_sources = (
            custom_data_sources if custom_data_sources is not None else []
        )
        self.timestamp_keys = timestamp_keys if timestamp_keys is not None else []
        self._tags = tags

    # @property
    # @deprecated("FeatureTable.primary_keys", since="v0.3.6")
    # def keys(self):
    #     return self.primary_keys

    @property
    def tags(self) -> Dict[str, str]:
        """
        Get the tags associated with the feature table.

        :return a Dictionary of all tags associated with the feature table as key/value pairs
        """
        if self._tags is None:
            # If no tags are set, self._tags is expected an empty dictionary.
            raise ValueError(
                "Internal error: tags have not been fetched for this FeatureTable instance"
            )
        return self._tags


    @classmethod
    def from_uc_get_table_response(cls, uc_get_table_response: Dict[str, object]):
        """Return a FeatureStore object from a UC get_table response. Note: UC does not return online_stores or tags.

        :param dict uc_get_table_response: A dictionary representing a UC get_table response.
        :return FeatureTable: a FeatureStore object from the UC response.
        """
        table_name = uc_get_table_response["full_name"]

        if uc_get_table_response["securable_kind"] == "TABLE_ONLINE_VIEW":
            source_table = uc_get_table_response["properties_pairs"]["properties"][
                "source_table"
            ]
            raise ValueError(
                f"Table '{table_name}' is an online view. Online Views are not feature tables. Please use the source table '{source_table}' instead."
            )

        if (
            "table_type" in uc_get_table_response
            and uc_get_table_response["table_type"] == "VIEW"
        ):
            return cls(
                name=table_name,
                table_id=uc_get_table_response["table_id"],
                description=uc_get_table_response["comment"]
                if "comment" in uc_get_table_response
                else "",
                primary_keys=[],
                partition_columns=[],
                features=[],
                creation_timestamp=uc_get_table_response["created_at"],
                timestamp_keys=[],
            )

        table_constraints = (
            uc_get_table_response["table_constraints"]
            if "table_constraints" in uc_get_table_response
            else []
        )
        primary_key_constraints = [
            c for c in table_constraints if "primary_key_constraint" in c
        ]
        if len(primary_key_constraints) == 0:
            raise ValueError(
                "Table can't be used as a feature table because it has no primary key constraint defined."
                + " Use 'ALTER TABLE table_name ADD CONSTRAINT table_name_pk PRIMARY KEY( key_column [,...] )'"
                + " to add a primary key constraint on the table."
            )
        primary_key_constraint = primary_key_constraint = primary_key_constraints[0][
            "primary_key_constraint"
        ]
        timestamp_keys = (
            primary_key_constraint["timeseries_columns"]
            if "timeseries_columns" in primary_key_constraint
            else []
        )
        primary_keys = [
            c
            for c in primary_key_constraint["child_columns"]
            if c not in timestamp_keys
        ]

        columns = uc_get_table_response["columns"]
        features = [c["name"] for c in columns]
        partition_columns_unordered = [c for c in columns if "partition_index" in c]
        partition_columns = [
            c["name"]
            for c in sorted(
                partition_columns_unordered, key=lambda x: x["partition_index"]
            )
        ]

        return cls(
            name=table_name,
            table_id=uc_get_table_response["table_id"],
            description=uc_get_table_response["comment"]
            if "comment" in uc_get_table_response
            else "",
            primary_keys=primary_keys,
            partition_columns=partition_columns,
            features=features,
            creation_timestamp=uc_get_table_response["created_at"],
            timestamp_keys=timestamp_keys,
        )



import json
import logging
import os
from collections import defaultdict
from types import ModuleType
from typing import Any, Dict, List, Optional, Set, Union

import mlflow
import yaml
from mlflow.models import Model, ModelSignature
from mlflow.utils.file_utils import TempDir, YamlSafeDumper, read_yaml
from pyspark.sql import DataFrame
from pyspark.sql.functions import struct

from feature_store.entities.feature_column_info import FeatureColumnInfo
from feature_store.entities.feature_function import FeatureFunction
from feature_store.entities.feature_lookup import FeatureLookup
from feature_store.entities.feature_spec import FeatureSpec
from feature_store.entities.training_set import TrainingSet
from feature_store.spark_client.spark_client import SparkClient

from feature_store.constants.constants import (
    _NO_RESULT_TYPE_PASSED,
    _PREBUILT_ENV_URI,
    _USE_SPARK_NATIVE_JOIN,
    _WARN,
    MODEL_DATA_PATH_ROOT,
    PREDICTION_COLUMN_NAME,
)

from feature_store.utils import common_utils, training_set_utils
from feature_store.utils.feature_spec_utils import convert_to_yaml_string

_logger = logging.getLogger(__name__)

FEATURE_SPEC_GRAPH_MAX_COLUMN_INFO = 1000


class TrainingSetClient:
    def __init__(
        self,
        spark_client: SparkClient
    ):
        self._spark_client = spark_client

    def create_training_set(
        self,
        feature_spec: FeatureSpec,
        feature_column_infos: List[FeatureColumnInfo],
        label_names: List[str],
        df: DataFrame,
        ft_metadata: training_set_utils._FeatureTableMetadata,
        kwargs,
    ):
        uc_function_infos = training_set_utils.get_uc_function_infos(
            self._spark_client,
            {odci.udf_name for odci in feature_spec.on_demand_column_infos},
        )

        # TODO(divyagupta-db): Move validation from _validate_join_feature_data in feature_lookup_utils.py
        #  to a helper function called here and in score_batch.

        # Add consumer of each feature and instrument as final step
        consumer_feature_table_map = defaultdict(list)
        for feature in feature_column_infos:
            consumer_feature_table_map[feature.table_name].append(feature.feature_name)
        consumed_udf_names = [f.udf_name for f in feature_spec.function_infos]

        # Spark query planning is known to cause spark driver to crash if there are many feature tables to PiT join.
        # See https://docs.google.com/document/d/1EyA4vvlWikTJMeinsLkxmRAVNlXoF1eqoZElOdqlWyY/edit
        # So we disable native join by default.
        training_set_utils.warn_if_non_photon_for_native_spark(
            kwargs.get(_USE_SPARK_NATIVE_JOIN, False), self._spark_client
        )
        return TrainingSet(
            feature_spec,
            df,
            label_names,
            ft_metadata.feature_table_metadata_map,
            ft_metadata.feature_table_data_map,
            uc_function_infos,
            kwargs.get(_USE_SPARK_NATIVE_JOIN, False),
        )

    def create_training_set_from_feature_lookups(
        self,
        df: DataFrame,
        feature_lookups: List[Union[FeatureLookup, FeatureFunction]],
        label: Union[str, List[str], None],
        exclude_columns: List[str],
        **kwargs,
    ) -> TrainingSet:

        # 获取特征查找列表和特征函数列表
        features = feature_lookups
        feature_lookups = [f for f in features if isinstance(f, FeatureLookup)]
        feature_functions = [f for f in features if isinstance(f, FeatureFunction)]

        # 如果未提供标签，则用空列表初始化label_names
        label_names = common_utils.as_list(label, [])
        del label

        # 校验数据集和标签
        training_set_utils.verify_df_and_labels(df, label_names, exclude_columns)

        # 获取特征表元数据
        ft_metadata = training_set_utils.get_table_metadata(
            self._spark_client,
            {fl.table_name for fl in feature_lookups}
        )

        column_infos = training_set_utils.get_column_infos(
            feature_lookups,
            feature_functions,
            ft_metadata,
            df_columns=df.columns,
            label_names=label_names,
        )

        training_set_utils.validate_column_infos(
            self._spark_client,
            ft_metadata,
            column_infos.source_data_column_infos,
            column_infos.feature_column_infos,
            column_infos.on_demand_column_infos,
            label_names,
        )

        # Build feature_spec locally for comparison with the feature spec yaml generated by the
        # FeatureStore backend. This will be removed once the migration is validated.
        feature_spec = training_set_utils.build_feature_spec(
            feature_lookups,
            ft_metadata,
            column_infos,
            exclude_columns
        )

        return self.create_training_set(
            feature_spec,
            column_infos.feature_column_infos,
            label_names,
            df,
            ft_metadata,
            kwargs=kwargs,
        )





    def create_feature_spec(
        self,
        name: str,
        features: List[Union[FeatureLookup, FeatureFunction]],
        sparkClient: SparkClient,
        exclude_columns: List[str] = [],
    ) -> FeatureSpec:

        feature_lookups = [f for f in features if isinstance(f, FeatureLookup)]
        feature_functions = [f for f in features if isinstance(f, FeatureFunction)]

        # Maximum of 100 FeatureFunctions is supported
        if len(feature_functions) > training_set_utils.MAX_FEATURE_FUNCTIONS:
            raise ValueError(
                f"A maximum of {training_set_utils.MAX_FEATURE_FUNCTIONS} FeatureFunctions are supported."
            )

        # Get feature table metadata and column infos
        ft_metadata = training_set_utils.get_table_metadata(
            self._spark_client,
            {fl.table_name for fl in feature_lookups}
        )
        column_infos = training_set_utils.get_column_infos(
            feature_lookups,
            feature_functions,
            ft_metadata,
        )

        column_infos = training_set_utils.add_inferred_source_columns(column_infos)

        training_set_utils.validate_column_infos(
            self._spark_client,
            ft_metadata,
            column_infos.source_data_column_infos,
            column_infos.feature_column_infos,
            column_infos.on_demand_column_infos,
        )

        feature_spec = training_set_utils.build_feature_spec(
            feature_lookups,
            ft_metadata,
            column_infos,
            exclude_columns
        )

        return feature_spec



from setuptools import setup, find_packages

setup(
    name="wedata_feature_engineering",
    version="0.1.2",
    packages=find_packages(exclude=["tests*"]),
    install_requires=[
        'pyspark>=3.0.0',
        'delta-spark>=1.0.0',
        'pandas>=1.0.0'
    ],
    python_requires='>=3.7',
    author="meahqian",
    author_email="",
    description="Wedata Feature Engineering Library",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    license="Apache 2.0",
    url="",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
)
